#include "xwatch.h"
#include "gui.h"

// global & option variables
int make_scrollbar_h = FALSE;
int make_scrollbar_v = TRUE;
struct Watch *first_watch = NULL;
int rc = 1;
char *geometry = NULL;

void
output_line (watch, str, len)
  struct Watch *watch;
  char *str;
  int len;
{
	struct tm linetime;

	if (len >= 16 && watch->flags & WATCH_FLAG_SYSLOG) {
		int c;
		syslog_time_to_tm(str, &linetime);
		for (c = 16; str[c] && !isspace(str[c]); c++); /* skip first word (computer name) */
		for (; str[c] && isspace(str[c]); c++); /* skip whitespace after that */
		if (!str[c]) return;
		str = &str[c];
		len -= c;
	} else {
		time_t secs;
		if (time(&secs) == -1) watch->flags &= ~WATCH_FLAG_TIMESTAMP;	/* stop timestamping of time fails */
		else memcpy(&linetime, localtime(&secs), sizeof(struct tm));
	}
	if (watch->flags & WATCH_FLAG_TIMESTAMP) {
		int tlen;
		char timestr[100];
		if (watch->timestamp) tlen = strftime(timestr, 100, watch->timestamp, &linetime);
		else tlen = strftime(timestr, 100, "%H:%M:%S ", &linetime);
		if (tlen > 0) print_line(watch, timestr, tlen, watch->ts_font, watch->ts_fgcolor, watch->ts_bgcolor);
	}

	print_line(watch, str, len, watch->font, watch->fgcolor, watch->bgcolor);
}

/* update_watch:
 *   read and print data from a watch
 */
#define READ_BUFFER_SIZE 100
void
update_watch (watch)
  struct Watch *watch;
{
	char buf[READ_BUFFER_SIZE+1];
	char *inputbuf;
	ssize_t len;

	do {
		len = read(watch->fd, buf, READ_BUFFER_SIZE);
		if (len == -1) {
			if (errno == EAGAIN || errno == EWOULDBLOCK) break;
#if HAVE_ASPRINTF
			len = asprintf(&inputbuf, "%sRemoving watch on %s (%s)\n", "xwatch: ", watch->filename, strerror(errno));
#else
			inputbuf = malloc(31+strlen(watch->filename)+strlen(strerror(errno)));
			len = sprintf(inputbuf, "%sRemoving watch on %s (%s)\n", "xwatch: ", watch->filename, strerror(errno));
#endif
			output_line(watch, inputbuf, len);
			free(inputbuf);
			close_watch(watch);
		} else if (len > 0) {
			buf[len] = 0;
			inputbuf = buf;
			if (watch->flags) { /* must read whole line if parsing is enabled */
				char *line;
				while ((line = get_whole_line(&inputbuf)))
					print_pending_and_line(watch, line);
				append_to_pending(watch, inputbuf);
			} else {
				output_line(watch, buf, strlen(buf));
			}
		}
	} while (len > 0);
}

int
open_file_watch (watch, filename)
  struct Watch *watch;
  char *filename;
{
	watch->id = 0;
	watch->pending = NULL;
	watch->pendinglen = 0;

	if (filename[0] == '-' && !filename[1]) {
		int flags;
		watch->fd = STDIN_FILENO;
		watch->type = WATCH_FIFO;
		fcntl(STDIN_FILENO, F_GETFL, &flags);
		flags |= O_NONBLOCK;
		fcntl(STDIN_FILENO, F_SETFL, flags);
	} else {
		struct stat statbuf;
		watch->fd = open(filename, O_RDONLY|O_NONBLOCK); // |O_NONBLOCK
		if (watch->fd == -1) { printerror(filename, ": ", strerror(errno), NULL); return FALSE; }
		if (fstat(watch->fd, &statbuf) == -1) { printerror(filename, ": ", strerror(errno), NULL); return FALSE; }
		if (S_ISREG(statbuf.st_mode)) watch->type = WATCH_REGULAR;
		else if (S_ISFIFO(statbuf.st_mode)) watch->type = WATCH_FIFO;
		else if (S_ISCHR(statbuf.st_mode)) watch->type = WATCH_FIFO;
		else if (S_ISSOCK(statbuf.st_mode)) watch->type = WATCH_SOCKET;
		else { printerror(filename, ": cannot watch files of that type", NULL); return FALSE; }
	}
	if (watch->type == WATCH_REGULAR) {
		if (lseek(watch->fd, 0, SEEK_END) == -1) { printerror(filename, ": ", strerror(errno), NULL); return FALSE; }
		watch->id = (void *) monitor_watch_timed(watch);
	} else if (watch->type == WATCH_FIFO) {
		watch->id = (void *) monitor_watch_event(watch);
	}

	return TRUE;
}

char *
get_arg_word (argc, argv, filename)
  int argc;
  char **argv;
  char *filename;
{
	static char *configstr = NULL;
	static int configsize;
	static int c = 1;
	static char *word;

	if (argv) {
		word = argv[c];
		if (!word) return NULL;
		c++;
	}
	else {
		int c2 = 0;
		if (!configstr) { /* first time */
			struct stat statbuf;
			int fd;
			fd = open(filename, O_RDONLY);
			if (fd == -1) { printerror(filename, ": ", strerror(errno), NULL); return (void *)1; }
			if (fstat(fd, &statbuf) == -1) { printerror(filename, ": ", strerror(errno), NULL); return (void *)1; }
			configsize = statbuf.st_size;
			if (!(configstr = malloc(configsize + 1))) return (void *)1;
			if (!xread(fd, configstr, configsize)) { printerror(filename, ": ", strerror(errno), NULL); return (void *)1; }
			close(fd);
			configstr[configsize] = 0;
			word = configstr;
		} else {
			while (word[c2]) c2++; /* skip to next null byte */
			if (&word[c2] == &configstr[configsize]) return NULL; /* all words parsed */
			c2++; /* skip past null byte */
		}
		while (isspace(word[c2])) c2++; /* skip whitespace characters after null byte */
		if (word[c2] == '"') {	/* quoted word */
			word = &word[c2+1];
			c2 = 0;
			while (word[c2] && word[c2] != '"') c2++;
		} else { /* standard space separated word */
			word = &word[c2];
			c2 = 0;
			while (word[c2] && !isspace(word[c2])) c2++;
		}
		if (&word[c2] == &configstr[configsize]) return NULL; /* all words parsed */
		word[c2] = 0;
	}

	return word;
}

#define VERSION_MESSAGE "xwatch 0.1.0\n"
#define HELP_MESSAGE "\
Usage: xwatch [OPTION]... [FILE]...\n\
Display the output of files (regular, fifos, stdin etc) in an X window.\n\
Order of arguments are important - color and font options affect the\n\
following file or timestamp arguments.\n\
\n\
  -f, --file=FILE           add watch for a file (default option)\n\
      --timestamp[=FORMAT]  display time for every line read\n\
                            (see strftime(3) for format, default is \"%H:%M:%S\")\n\
      --no-timestamp        disable timestamping\n\
      --fgcolor=COLOR       foreground color, in hexadecimal format \"RRGGBB\"\n\
      --no-fgcolor          use built-in default color as foreground\n\
      --bgcolor=COLOR       background color, in hexadecimal format \"RRGGBB\"\n\
      --no-bgcolor          use built-in default color as background\n\
      --font=FONT           font, in standard X font specification\n\
      --no-font             use built-in default font\n\
  -C, --config=FILE         read options from specified file\n\
  -d  --delay=DELAY         delay in milliseconds between reads (regular files)\n\
      --syslog              enable syslog parsing\n\
      --no-syslog           display syslog parsing (default)\n\
      --hscrollbar          enable horizontal scrollbar\n\
      --no-hscrollbar       disable horizontal scrollbar (default)\n\
      --vscrollbar          enable vertical scrollbar (default)\n\
      --no-vscrollbar       disable vertical scrollbar\n\
      --help                display this help and exit\n\
      --version             output version information and exit\n\
\n\
Report bugs to osk@hem.passagen.se\n"

//      --geometry            geometry of window (see X(1) for format)\n\


int
parse_args (argc, argv, filename)
  int argc;
  char **argv;
  char *filename;
{
	guint current_delay = 2000;			/* read delay in milliseconds */
	char *current_fgcolor = NULL;		/* text foreground color */
	char *current_bgcolor = NULL;		/* text background color */
	char *current_ts_fgcolor = NULL;	/* timestamp foreground color */
	char *current_ts_bgcolor = NULL;	/* timestamp background color */
	char *current_ts_font = NULL;		/* timestamp font */
	char *current_font = NULL;			/* text font */
	char *current_timestamp = NULL;		/* timestamp string */
	int current_flags = 0;				/* watch parsing flags */

	int pending_word = XWATCH_OPT_FILE;	/* type of pending argument word */
	int force_word = 0;					/* force word even if it looks like an argument (-1=always, 1=next word only) */
	struct Watch *watch;
	char *word;

	int res = FALSE;

	while (TRUE) {
		word = get_arg_word(argc, argv, filename);	/* get next word */
		if (!word) { res = TRUE; break; }
		else if (word == (void *) 1) break;
		//printf("word=%s\n", word);

		if (!force_word) {
			if (word[0] == '-' && word[1] == '-') {	/* long option */
				char *argvalue, *argend;
			 	if (!word[2]) { force_word = -1; continue; }
				if ((argvalue = index(word, '='))) {
					if (!argc) {
						argend = index(word, '\0');
						argend[0] = ' ';	/* shameless fix */
					}
					argvalue[0] = 0;
					force_word = 1;
				}
				if (!strcmp(&(word[2]), "file")) pending_word = XWATCH_OPT_FILE;
				else if (!strcmp(&(word[2]), "geometry")) pending_word = XWATCH_OPT_GEOMETRY;
				else if (!strcmp(&(word[2]), "help")) {
					write(STDOUT_FILENO, HELP_MESSAGE, strlen(HELP_MESSAGE));
					rc = 0; res = FALSE;
					break;
				}
				else if (!strcmp(&(word[2]), "version")) {
					write(STDOUT_FILENO, VERSION_MESSAGE, strlen(VERSION_MESSAGE));
					rc = 0; res = FALSE;
					break;
				}
				else if (!strcmp(&(word[2]), "font")) pending_word = XWATCH_OPT_FONT;
				else if (!strcmp(&(word[2]), "config")) pending_word = XWATCH_OPT_CONFIG;
				else if (!strcmp(&(word[2]), "delay")) pending_word = XWATCH_OPT_DELAY;
				else if (!strcmp(&(word[2]), "syslog")) current_flags |= WATCH_FLAG_SYSLOG;
				else if (!strcmp(&(word[2]), "no-syslog")) current_flags &= ~WATCH_FLAG_SYSLOG;
				else if (!strcmp(&(word[2]), "timestamp")) {
					if (argvalue) pending_word = XWATCH_OPT_TIMESTAMP;
					current_flags |= WATCH_FLAG_TIMESTAMP;
					if (current_fgcolor) {
						if (!(current_ts_fgcolor = strdup(current_fgcolor))) break;
						else current_ts_fgcolor = NULL;
					}
					if (current_bgcolor) {
						if (!(current_ts_bgcolor = strdup(current_bgcolor))) break;
						else current_ts_bgcolor = NULL;
					}
					if (current_font) {
						if (!(current_ts_font = strdup(current_font))) break;
						else current_ts_font = NULL;
					}
				}
				else if (!strcmp(&(word[2]), "no-timestamp")) {
					current_flags &= ~ WATCH_FLAG_TIMESTAMP;
				    if (current_timestamp) { free(current_timestamp); current_timestamp = NULL; }
				    if (current_ts_fgcolor) { free(current_ts_fgcolor); current_ts_fgcolor = NULL; }
				    if (current_ts_bgcolor) { free(current_ts_bgcolor); current_ts_bgcolor = NULL; }
				    if (current_ts_font) { free(current_ts_font); current_ts_font = NULL; }
				}
				else if (!strcmp(&(word[2]), "hscrollbar")) make_scrollbar_h = TRUE;
				else if (!strcmp(&(word[2]), "no-hscrollbar")) make_scrollbar_h = FALSE;
				else if (!strcmp(&(word[2]), "vscrollbar")) make_scrollbar_v = TRUE;
				else if (!strcmp(&(word[2]), "no-vscrollbar")) make_scrollbar_v = FALSE;
#ifdef US_ENGLISH
				else if (!strcmp(&(word[2]), "fgcolor")) pending_word = XWATCH_OPT_FGCOLOR;
				else if (!strcmp(&(word[2]), "bgcolor")) pending_word = XWATCH_OPT_BGCOLOR;
#endif /* US_ENGLISH */
#ifdef UK_ENGLISH
				else if (!strcmp(&(word[2]), "fgcolour")) pending_word = XWATCH_OPT_FGCOLOR;
				else if (!strcmp(&(word[2]), "bgcolour")) pending_word = XWATCH_OPT_BGCOLOR;
#endif /* UK_ENGLISH */
				else { printerror("unrecognized option `", word, "'", NULL); break; }
				if (!(argvalue && argc))
					continue;
				else
					word = &argvalue[1];
			} else if (word[0] == '-' && word[1]) { /* short options */
				register int c2;
				char dummy[2];
				for(c2 = 1; word[c2]; c2++) {
					if (word[c2] == 'f') pending_word = XWATCH_OPT_FILE;
					else if (word[c2] == 'C') pending_word = XWATCH_OPT_CONFIG;
					else if (word[c2] == 'd') pending_word = XWATCH_OPT_DELAY;
					else {
						dummy[0] = word[c2]; dummy[1] = 0;
						printerror("unrecognized option `-", dummy, "'", NULL);
						c2 = -1; break;
					}
				}
				if (c2 == -1) break;
				continue;
			}
		}

		if (pending_word == XWATCH_OPT_CONFIG)
			parse_args(0, NULL, word);
		else if (pending_word == XWATCH_OPT_DELAY)
			current_delay = atoi(word);
		else if (pending_word == XWATCH_OPT_FONT) {
			if (current_font) free(current_font);
			if (!(current_font = strdup(word))) break;
		}
		else if (pending_word == XWATCH_OPT_FGCOLOR) {
			if (current_fgcolor) free(current_fgcolor);
			if (!(current_fgcolor = strdup(word))) break;
		}
		else if (pending_word == XWATCH_OPT_BGCOLOR) {
			if (current_bgcolor) free(current_bgcolor);
			if (!(current_bgcolor = strdup(word))) break;
		}
		else if (pending_word == XWATCH_OPT_TIMESTAMP) {
			if (current_timestamp) free(current_timestamp);
			if (!(current_timestamp = strdup(word))) break;
		}
		else if (pending_word == XWATCH_OPT_GEOMETRY) {
			if (geometry) free(geometry);
			if (!(geometry = strdup(word))) break;
		}
		else {
			watch = malloc(sizeof(struct Watch));
			if (first_watch) watch->next = first_watch;
			else watch->next = NULL;
			first_watch = watch;
			watch->delay = current_delay;
			watch->flags = current_flags;
			watch->fgcolor = watch->bgcolor = watch->ts_fgcolor = watch->ts_bgcolor = watch->ts_font = watch->timestamp = watch->filename = NULL;
			if (current_font && !(watch->font = load_font(current_font))) { printerror("could not load font `", current_font, "'", NULL); break; }
			if (current_fgcolor && !(watch->fgcolor = load_color(current_fgcolor))) { printerror("could not load color `", current_fgcolor, "'", NULL); break; }
			if (current_bgcolor && !(watch->bgcolor = load_color(current_bgcolor))) { printerror("could not load color `", current_bgcolor, "'", NULL); break; }
			if (current_ts_font && !(watch->ts_font = load_font(current_ts_font))) { printerror("could not load font `", current_ts_font, "'", NULL); break; }
			if (current_ts_fgcolor && !(watch->ts_fgcolor = load_color(current_ts_fgcolor))) { printerror("could not load color `", current_ts_fgcolor, "'", NULL); break; }
			if (current_ts_bgcolor && !(watch->ts_bgcolor = load_color(current_ts_bgcolor))) { printerror("could not load color `", current_ts_bgcolor, "'", NULL); break; }
			if (current_timestamp && !(watch->timestamp = strdup(current_timestamp))) break;
			if (word[0] == '-' && word[1] == 0) { watch->filename = strdup("standard output"); }
			else { watch->filename = strdup(word); }
			if (!watch->filename) break;
			if (!open_file_watch(watch, word)) break;
		}
		pending_word = XWATCH_OPT_FILE;
		if (force_word > 0) force_word--;
	}
	if (current_timestamp) free(current_timestamp);
	if (current_font) free(current_font);
	if (current_ts_font) free(current_ts_font);
	if (current_fgcolor) free(current_fgcolor);
	if (current_bgcolor) free(current_bgcolor);
	if (current_ts_fgcolor) free(current_ts_fgcolor);
	if (current_ts_bgcolor) free(current_ts_bgcolor);

	return res;
}

void
close_watches (void)
{
	struct Watch *watch, *nextwatch;

	watch = first_watch;
	while (watch) {
		nextwatch = watch->next;
		close_watch(watch);
		watch = nextwatch;
	}
}

void
close_watch (watch)
  struct Watch *watch;
{
	struct Watch *tmpwatch;

	/* unlink from list */
	tmpwatch = first_watch;
	if (first_watch == watch) first_watch = watch->next;
	else {
		while (tmpwatch) {
			if (tmpwatch->next == watch) { tmpwatch->next = watch->next; break; }
			tmpwatch = tmpwatch->next;
		}
	}
	if (watch->id) {
		if (watch->type & WATCH_TYPE_EVENT) unmonitor_watch_event(watch);
		else unmonitor_watch_timed(watch);
	}
	if (watch->fd > STDERR_FILENO) close(watch->fd);
	if (watch->font) unload_font(watch->font);
	if (watch->ts_font) unload_font(watch->ts_font);
	if (watch->fgcolor) unload_color(watch->fgcolor);
	if (watch->bgcolor) unload_color(watch->bgcolor);
	if (watch->ts_fgcolor) unload_color(watch->ts_fgcolor);
	if (watch->ts_bgcolor) unload_color(watch->ts_bgcolor);
	if (watch->timestamp) free(watch->timestamp);
	free(watch->filename);
	free(watch);
}

RETSIGTYPE
terminate_signal(sig)
  int sig;
{
	close_watches();
	exit_gui(0);
}

int
main (argc, argv)
  int argc;
  char **argv;
{
	struct sigaction action;
	sigset_t set;

	/* initialize gui */
	if (!init_gui(&argc, &argv))
		return rc;

	/* install signal handlers */
	action.sa_handler = SIG_IGN;
	sigemptyset(&set);
	action.sa_mask = set;
	action.sa_flags = SA_RESTART;
	if (sigaction(SIGPIPE, &action, NULL)
	    || sigaction(SIGHUP, &action, NULL)) {
		printerror(strerror(errno), NULL);
		exit_gui(rc);
	}

	action.sa_handler = terminate_signal;
	sigaddset(&set, SIGTERM);
	sigaddset(&set, SIGUSR1);
	if (sigaction(SIGTERM, &action, NULL)) {
		printerror(strerror(errno), NULL);
		exit_gui(rc);
	}


	/* parse command line arguments */
	if (!parse_args(argc, argv, NULL)) {
		close_watches();
		exit_gui(rc);
	}
	
	/* start the application and the gui */
	if (!start_gui())
		exit_gui(rc);

	exit_gui(0);

	return 0;
}
